--[==[

Copyright ©2019 Samuel Thomas Pain

The contents of this addon, excluding third-party resources, are
copyrighted to their authors with all rights reserved.

This addon is free to use and the authors hereby grants you the following rights:

1.     You may make modifications to this addon for private use only, you
    may not publicize any portion of this addon.

2.     Do not modify the name of this addon, including the addon folders.

3.     This copyright notice shall be included in all copies or substantial
    portions of the Software.

All rights not explicitly addressed in this license are reserved by
the copyright holders.

]==]--


local ADDON_NAME, GHC = ...

local DEBUG = false

local addon = LibStub("AceAddon-3.0"):NewAddon(GHC.addon_name, "AceConsole-3.0", "AceEvent-3.0", "AceComm-3.0", "AceSerializer-3.0")
addon.Name = ADDON_NAME
addon.LocName = select(2, GetAddOnInfo(addon.Name))
addon.Notes = select(3, GetAddOnInfo(addon.Name))

local tinsert, tconcat, tremove, tsort = table.insert, table.concat, table.remove, table.sort

GHC.Functions = {}

GHC.Pending = nil

GHC.Var = {
    NewVersion = "0.1.0",
    RosterSort = nil,
    MainsName = "-",
}

local IsGuildspaceRegistered = false


-------------------------------------------------------------------------------
-- HELPER FUNCTIONS
-------------------------------------------------------------------------------

--- Deep copy of all entries in `...` sources to `dest`
-- Entries in `dest` will be overwritten by entries in the `...` sources if they have the same key.
-- Later sources' entries will overwrite earlier ones.
--
-- WARNING: assign({a = 1}, "erase") == "erase"
local function assign(dest, ...)
    for i = 1, select("#", ...) do
        local varg = select(i, ...)
        if type(varg) == "table" then
            for k, v in pairs(varg) do
                dest[k] = assign(dest[k] or {}, v)
            end
        else  -- number, string, boolean, etc.
            dest = varg
        end
    end
    return dest
end


--- Return a table of keys sorted by `f` that can be used to index into the passed table
local function getSortedKeysByValue(tbl, sortFunction)
    local keys = {}
    for k in pairs(tbl) do
        tinsert(keys, k)
    end
    tsort(keys, function(a, b) return sortFunction(tbl[a], tbl[b]) end)
    return keys
end


local function createMemberTable(...)
    local member = {
        Name = nil,
        Level = nil,
        Class = nil,
        Gender = nil,
        Race = nil,
        Professions = {
            Primary = {
                -- { Name = nil, Level = nil, },
                -- { Name = nil, Level = nil, },
            },
        },
        Role = nil,
        PublicNote = nil,
        MainsName = nil,
    }
    return assign(member, ...)
end


local rosterSortFunctions = {
    Name  = function(a, b) return a.Name  < b.Name  end,  -- ascending
    Class = function(a, b) return a.Class < b.Class end,  -- ascending
    Level = function(a, b) return a.Level > b.Level end,  -- descending
    Role  = function(a, b) return a.Role  > b.Role  end,  -- descending
}


local function tprint(tbl)
    local ret = {}
    for k, v in pairs(tbl) do
        if type(v) == "table" then
            tinsert(ret, ("%s={%s}"):format(k, tprint(v)))
        else
            tinsert(ret, ("%s=%s"):format(k, v))
        end
    end
    return tconcat(ret, ", ")
end


local function keyCount(tbl)
    local i = 0
    for _ in pairs(tbl) do i = i + 1 end
    return i
end


local function updateListViewItem(listViewItem, member)
    -- Name, MainsName: set to "-" if nil
    if member.Name then
        listViewItem.Name:SetText(
            ("%s [|cffABD473%s|r]"):format(
                member.Name,
                member.MainsName or "-"
            )
        )
    else
        listViewItem.Name:SetText("")
    end

    -- Level
    listViewItem.Level:SetText(member.Level or "")

    -- ClassIcon
    if member.Class then
        listViewItem.ClassIcon.Texture:SetTexture(
            "Interface/Addons/GuildHelperClassic/ClassIcons/" .. member.Class
        )
    else
        listViewItem.ClassIcon.Texture:SetTexture(nil)
    end

    -- RaceIcon
    if member.Gender and member.Race then
        listViewItem.RaceIcon.Texture:SetTexture(GHC.Db.RaceIcons[member.Gender][member.Race])
    else
        listViewItem.RaceIcon.Texture:SetTexture(nil)
    end

    -- RoleText
    if member.Role then
        -- Default to 0 if Role is undefined
        listViewItem.RoleText:SetText(GHC.Db.RoleIconTextString[tonumber(member.Role) or 0])
    else
        listViewItem.RoleText:SetText("")
    end

    local primaries = member.Professions and member.Professions.Primary or {}
    for i, professionFrame in ipairs(listViewItem.Professions) do
        local profession = primaries[i]
        if profession then
            professionFrame.Icon:Show()
            professionFrame.Icon.Texture:SetTexture(GHC.Db.ProfessionIcons[profession.Name])
            professionFrame.Text = ("%s [%d]"):format(profession.Name, profession.Level)
        else
            professionFrame.Icon:Hide()
            professionFrame.Icon.Texture:SetTexture(nil)
            professionFrame.Text = ""
        end
    end
end


--- Debug function to find event args
function GHC.Functions.GetArgs(...)
    for i = 1, select("#", ...) do
        arg = select(i, ...)
        print(i.." "..tostring(arg))
    end
end


local function migrateOldGuildspace(guildName)
    local newGuildspace = addon.db.factionrealm[guildName]
    if not newGuildspace then return end

    -- Attempt to migrate old data
    if GhcDb then
        local oldGuildspace = GhcDb[guildName]
        if oldGuildspace then
            addon:Print("Upgrading data for", guildName)
            if oldGuildspace.GuildDb then
                for _, member in ipairs(oldGuildspace.GuildDb) do
                    -- name, level class, public note can be retrieved during ScanGuildRoster
                    newGuildspace.Roster[member.GUID] = createMemberTable({
                        Race = member.Race,
                        Gender = member.Gender,
                        Role = tonumber(member.Role),
                        MainsName = member.MainsName,
                    })
                    local p = 1
                    if member.Profession1Name and member.Profession1Level then
                        newGuildspace.Roster[member.GUID].Professions.Primary[p] = {
                            Name = member.Profession1Name,
                            Level = member.Profession1Level,
                        }
                        p = p + 1
                    end
                    if member.Profession2Name and member.Profession2Level then
                        newGuildspace.Roster[member.GUID].Professions.Primary[p] = {
                            Name = member.Profession2Name,
                            Level = member.Profession2Level,
                        }
                        p = p + 1
                    end
                end
            end
            if oldGuildspace.GuildBank then
                newGuildspace.Bank = assign({}, oldGuildspace.GuildBank)
            end
            if oldGuildspace.GuildBankMoney then
                newGuildspace.Money = oldGuildspace.GuildBankMoney
            end

            -- TODO Finally, wipe out the old data
            -- GhcDb[guildName] = nil
        end
    end
end


-------------------------------------------------------------------------------
-- GUI MANAGEMENT FUNCTIONS
-------------------------------------------------------------------------------

function GHC_OnLoad(self)
    self:SetMovable(true)
    self:EnableMouse(true)
    self:RegisterForDrag("LeftButton")
    self:SetScript("OnDragStart", self.StartMoving)
    self:SetScript("OnDragStop", self.StopMovingOrSizing)
end


--- Set a frame to be moveable
function GHC.Functions.MakeFrameMove(frame)
    frame:SetMovable(true)
    frame:EnableMouse(true)
    frame:RegisterForDrag("LeftButton")
    frame:SetScript("OnDragStart", frame.StartMoving)
    frame:SetScript("OnDragStop", frame.StopMovingOrSizing)
end


--- Used for roster professions icons
function GHC.Functions.ShowTooltip(frame, text)
    GHC.UI.Tooltip:SetParent(frame)
    GHC.UI.Tooltip.Text:SetText(text)
    local w = GHC.UI.Tooltip.Text:GetStringWidth()
    GHC.UI.Tooltip:SetSize(w + 24, 30)
    GHC.UI.Tooltip:SetPoint("TOPLEFT", - w * 1.25, 12)
    GHC.UI.Tooltip:Show()
end


--- Clear roster list view items starting from lBound up to uBound
-- @param[opt=1] lBound Lower bound of listview to clear
-- @param[opt=10] uBound Upper bound of listview to clear
function GHC.Functions.ClearGuildRosterListView(lBound, uBound)
    local emptyMember = createMemberTable()

    -- The loop will not excecute if lBound > uBound
    for i = (lBound or 1), (uBound or 10) do
        updateListViewItem(GHC.GuildRosterListViewItems[i], emptyMember)
    end
end


function GHC.Functions.ScrollByAmount(scrollBar, delta)
    local value = scrollBar:GetValue()
    addon:Debug(scrollBar:GetValue(), delta, value)
    scrollBar:SetValue(value - delta)
end


--- Scroll by three members at a time
function addon:GuildRosterScrollBarChanged(value, rosterTable)
    local guildspace = self:GetGuildspace()
    if not guildspace then return end

    local rosterKeys = GHC.Var.RosterKeysFiltered or guildspace.RosterKeysSorted

    rosterTable = rosterTable or guildspace.Roster

    local v = math.ceil(value)
    local lBound = 1 + 3 * (v - 1)
    local uBound = math.min(lBound + 9, #rosterKeys)
    local i = 1
    -- Use i to access the listview item and k to access the guild roster
    for k = lBound, uBound do
        local guid = rosterKeys[k]
        if guid then
            local member = rosterTable[guid]
            updateListViewItem(GHC.GuildRosterListViewItems[i], member)
            i = i + 1
        end
    end

    -- Only clear those remaining unchanged items
    GHC.Functions.ClearGuildRosterListView(i)
end
function GHC.Functions.GuildRosterScrollBarChanged(...) return addon:GuildRosterScrollBarChanged(...) end


-------------------------------------------------------------------------------
-- DATA MANAGEMENT FUNCTIONS
-------------------------------------------------------------------------------

--- Used as a nuclear option to wipe and restore default db
function GHC.Functions.Reset(guildName)
    local guildspace = addon:GetGuildspace()
    if not guildspace then
        addon:Print("Either you are not in a guild or the guild information is not available; please try opening the guild window and run reset again.")
        return
    end

    local gName = addon:GetPlayerGuildInfo()
    if guildName == gName then
        addon:Print("Found matching guild name, creating new settings")
        addon:ResetGuildspace(guildName)
        -- addon.db.factionrealm[guildName] = nil
        -- IsGuildspaceRegistered = false
        -- addon:RegisterGuildspace(guildName)
    else
        addon:Print("Unable to find matching guild, please make sure guild name entered corectly")
    end
end


-- TODO Consider making these Filter functions members of the Roster table
-- guildspace.Roster:FilterByRole(role)
-- guildspace.Roster:FilterByProfession(role)
-- guildspace.Roster:FilterBy(filterFunc)

--- Filter the guild roster by role
-- @param role Role to filter
function GHC.Functions.FilterByRole(role)
    addon:FilterRosterBy(function(member)
        return member.Role == role
    end)
end


--- Filter the guild roster by profession
-- @param prof Profession name to filter
function GHC.Functions.FilterByProfession(prof)
    local function filterFunc(member)
        for _, professions in pairs(member.Professions or {}) do
            for _, profession in ipairs(professions or {}) do
                if profession.Name == prof then
                    return true
                end
            end
        end
        return false
    end
    addon:FilterRosterBy(filterFunc)
end


local function updateRosterListView(roster, guids)
    -- Update listview scroll limits and scroll to the top
    GHC.UI.RosterScrollBar:SetMinMaxValues(1, math.max(math.ceil(#guids / 3), 1))
    GHC.UI.RosterScrollBar:SetValue(1)

    -- Update at most the first 10 members or all members if there are less than 10
    local c = 1
    for i = 1, math.min(#guids, 10) do
        updateListViewItem(GHC.GuildRosterListViewItems[i], roster[guids[i]])
        c = c + 1
    end

    -- Clear remaining icons & text from listview
    GHC.Functions.ClearGuildRosterListView(c)
end


--- Filter the GuildDb roster including only members for which the passed `filterFunc` returns true
-- @param[opt=all] filterFunc(member) Include members for which filterFunc(member) returns true
function addon:FilterRosterBy(filterFunc)
    local guildspace = self:GetGuildspace()
    if not guildspace then return end

    -- Apply filterFunc to the roster and only display filtered members
    local filteredRosterKeys = {}
    for _, guid in ipairs(guildspace.RosterKeysSorted) do
        if filterFunc(guildspace.Roster[guid]) then
            tinsert(filteredRosterKeys, guid)
        end
    end
    GHC.Var.RosterKeysFiltered = filteredRosterKeys

    updateRosterListView(guildspace.Roster, filteredRosterKeys)
end  -- addon:FilterRosterBy


--- Scans the guild roster, creates a backup db which is checked against to update player info after roster scan
function addon:ScanGuildRoster()
    local guildspace = self:GetGuildspace()
    if not guildspace then return end

    -- Make a copy of Roster to RosterBackup
    guildspace.RosterBackup = guildspace.Roster

    -- Wipe guild roster db
    guildspace.Roster = {}
    GHC.Var.RosterKeysFiltered = nil

    local totalMembers, _, _ = GetNumGuildMembers()

    for i = 1, totalMembers do
        -- name, rankName, rankIndex, level, classDisplayName, zone, publicNote, officerNote, isOnline, status, class, achievementPoints, achievementRank, isMobile, canSoR, repStanding, guid
        local name, _, _, level, _, _, publicNote, _, _, _, class, _, _, _, _, _, guid = GetGuildRosterInfo(i)
        name = strsplit("-", name, 2)

        local member = {
            Name = name,
            Level = level,
            Class = class,
            PublicNote = publicNote,
        }

        -- If we have a back up use it as a baseline, otherwise just fill as much as we know from member
        -- Default Role to 0 "Unknown"
        guildspace.Roster[guid] = assign({ Role = 0 }, guildspace.RosterBackup[guid] or {}, member)
    end

    guildspace.RosterKeysSorted = getSortedKeysByValue(
        guildspace.Roster, rosterSortFunctions[GHC.Var.RosterSort or "Name"]
    )

    -- if not GHC.Var.ShouldUpdateDisplay then return end
    GHC.Var.ShouldUpdateDisplay = false

    updateRosterListView(guildspace.Roster, guildspace.RosterKeysSorted)

    -- Update the class summary, reset counts first then loop db and make new count > sort > update UI
    for _, data in pairs(guildspace.ClassCount) do
        data.Count = 0
    end
    for _, member in pairs(guildspace.Roster) do
        guildspace.ClassCount[member.Class].Count = guildspace.ClassCount[member.Class].Count + 1
    end
    guildspace.ClassCountKeysSorted = getSortedKeysByValue(
        guildspace.ClassCount, function (a, b) return a.Count > b.Count end
    )
    for i, class in ipairs(guildspace.ClassCountKeysSorted) do
        local count = guildspace.ClassCount[class].Count
        local bar = GHC.UI.ClassSummaryBars[i]
        bar.StatusBar:SetValue((count / totalMembers) * 100.0)
        bar.StatusBar:SetStatusBarColor(GHC.Db.ClassColours[class].r, GHC.Db.ClassColours[class].g, GHC.Db.ClassColours[class].b)
        bar.Icon.Texture:SetTexture(("Interface/Addons/GuildHelperClassic/ClassIcons/%s"):format(class))
        bar.Text:SetText(("%d [%.1f%%]"):format(count, (count / totalMembers) * 100.0))
    end
end
function GHC.Functions.ScanGuildRoster(...) return addon:ScanGuildRoster(...) end


function addon:CreateMyDataString()
    local guid = UnitGUID("player")
    local _, class, _, race, gender, name, _ = GetPlayerInfoByGUID(guid)

    local mainsName, role
    local guildspace = self:GetGuildspace()
    if guildspace then 
        local player = guildspace.Roster[guid]
        if player then
            mainsName = player.MainsName
            role = player.Role
        end
    end

    local player = createMemberTable({
        MainsName = mainsName or "-",
        Role = role or 0,
        Name = name,
        Level = UnitLevel("player"),
        Class = string.upper(class),
        Gender = gender == 3 and "FEMALE" or "MALE",
        Race = string.upper(race),
    })

    -- TODO talent data isn't gathered at the moment and the intention is to avoid using it, instead players select their role(s) regardless of class/talents

    local p = 1
    for i = 1, GetNumSkillLines() do
        -- TODO Does this return a generic skill ID that doesn't need localization?
        local skill, _, _, level, _, _, _, _, _, _, _, _, _ = GetSkillLineInfo(i)
        -- TODO Couldn't this be easily extended to include Fishing, Cooking, First Aid?
        if GHC.Constants.Professions.Primary[skill] then
            player.Professions.Primary[p] = {
                Name = skill,
                Level = level,
            }
            p = p + 1
        end
    end

    return self:Serialize(guid, player)
end
function GHC.Functions.CreateMyDataString(...) return addon:CreateMyDataString(...) end


function addon:SendMyData()
    self:SendCommMessage("ghc-playerdata", self:CreateMyDataString(), "GUILD", nil, "BULK")
end
function GHC.Functions.SendMyData(...) return addon:SendMyData(...) end


function addon:ParsePlayerData(ds, sender)
    local guildspace = self:GetGuildspace()
    if not guildspace then return end

    local success, guid, player = self:Deserialize(ds)
    if success then
        guildspace.Roster[guid] = player

        -- TODO Don't need to rescan the entire roster, just re-sort and update the view basically
        self:ScanGuildRoster()
    else
        -- TODO Until update-me message is reworked, this is a hack -- alert once per sender per login
        GHC.Var.Senders = GHC.Var.Senders or {}
        if not GHC.Var.Senders[sender] then
            GHC.Var.Senders[sender] = true
            self:Print(("Unable to parse data from %s - do they have an old version?"):format(sender))
            self:Debug(ds)
        end
        -- self:Print("error deserializing:", guid)
        -- self:Print("Make sure", GHC.addon_name, "is up to date!")
    end
end
function GHC.Functions.ParsePlayerData(...) return addon:ParsePlayerData(...) end


function addon:GuildBankStart(money)
    local guildspace = self:GetGuildspace()
    if not guildspace then return end

    GHC.Pending = {
        Money = money,
    }
end
function GHC.Functions.GuildBankStart(...) return addon:GuildBankStart(...) end


function addon:ScanGuildBank()
    local _, _, rankIndex, _ = GetGuildInfo("player")
    if rankIndex == 0 or rankIndex == 1 then
        self:Debug("ghc-gbdata-s", tostring(GetMoney()), "GUILD")
        self:SendCommMessage("ghc-gbdata-s", tostring(GetMoney()), "GUILD")

        local ds = {}
        local c = 0

        -- Scan basic bank slots
        for i = 1, 28 do
            -- icon, itemCount, locked, quality, readable, lootable, itemLink, isFiltered, noValue, itemID
            local icon, itemCount, locked, quality, readable, lootable, itemLink, isFiltered, noValue, itemID = GetContainerItemInfo(-1, i)
            if itemLink then
                tinsert(ds, {ID = itemID, Count = itemCount})
                c = c + 1
            end
        end

        -- Scan additional bags in bank slots
        for bag = 5, 11 do
            for slot = 1, GetContainerNumSlots(bag) do
                local icon, itemCount, locked, quality, readable, lootable, itemLink, isFiltered, noValue, itemID = GetContainerItemInfo(bag, slot)
                if itemLink then
                    tinsert(ds, {ID = itemID, Count = itemCount})
                    c = c + 1
                end
            end
        end

        -- Scan character bags
        for bag = 0, 4 do
            for slot = 1, GetContainerNumSlots(bag) do
                local icon, itemCount, locked, quality, readable, lootable, itemLink, isFiltered, noValue, itemID = GetContainerItemInfo(bag, slot)
                if itemLink then
                    tinsert(ds, {ID = itemID, Count = itemCount})
                    c = c + 1
                end
            end
        end

        -- TODO Timestamp - could just be a second serialized value that old clients would ignore?
        local serialized = self:Serialize(ds, GetServerTime())
        self:Debug("ghc-gbdata", serialized, "GUILD")
        self:SendCommMessage("ghc-gbdata", serialized, "GUILD")

        self:Debug("ghc-gbdata-f", "blank", "GUILD")
        self:SendCommMessage("ghc-gbdata-f", "blank", "GUILD")
    end
end
function GHC.Functions.ScanGuildBank(...) return addon:ScanGuildBank(...) end


function addon:SendGuildBankData()
    -- Don't send data if in the middle of receiving new data
    if GHC.Pending then
        self:Print("New data is being received - please try again in a few seconds")
        return
    end

    local guildspace = self:GetGuildspace()
    if not guildspace then return end

    -- Don't send data if we don't have a valid timestamp
    if not guildspace.BankTimestamp then return end

    self:Debug("ghc-gbdata-s", tostring(guildspace.Money), "GUILD")
    self:SendCommMessage("ghc-gbdata-s", tostring(guildspace.Money), "GUILD")

    local tbl = {}
    for _, item in ipairs(guildspace.Bank) do
        tinsert(tbl, { ID = item.ID, Count = item.Count })
    end

    -- TODO Timestamp - could just be a second serialized value that old clients would ignore?
    -- local ds = self:Serialize(tbl, guildspace.BankTimestamp)
    local ds = self:Serialize(tbl, guildspace.BankTimestamp)

    self:Debug("ghc-gbdata", ds, "GUILD")
    self:SendCommMessage("ghc-gbdata", ds, "GUILD")

    self:Debug("ghc-gbdata-f", "blank", "GUILD")
    self:SendCommMessage("ghc-gbdata-f", "blank", "GUILD")
end
function GHC.Functions.SendGuildBankData(...) return addon:SendGuildBankData(...) end


--- Close guild bank parsing
function addon:GuildBankFinish()
    GHC.Pending = nil
end
function GHC.Functions.GuildBankFinish(...) return addon:GuildBankFinish(...) end


function addon:ParseGuildBankData(dataString, sender)
    if not dataString then return end
    if not GHC.Pending then return end

    local guildspace = self:GetGuildspace()
    if not guildspace then return end

    -- TODO Timestamp - could just be a second serialized value that old clients would ignore?
    local success, tbl, timestamp = self:Deserialize(dataString)
    if success then
        -- Timestamp: Ignore data without timestamp or with older/same timestamp
        self:Debug("Timestamp", timestamp, guildspace.BankTimestamp)
        if not timestamp then
            -- TODO Until update-me message is reworked, this is a hack -- alert once per sender per login
            GHC.Var.Senders = GHC.Var.Senders or {}
            if not GHC.Var.Senders[sender] then
                GHC.Var.Senders[sender] = true
                self:Print(("Unable to parse data from %s without a valid timestamp - do they have an old version?"):format(sender))
            end
            return
        end
        if guildspace.BankTimestamp and timestamp <= guildspace.BankTimestamp then return end

        guildspace.Money = GHC.Pending.Money

        -- Wipe old bank data
        guildspace.Bank = {}

        for i, item in ipairs(tbl) do
            -- Type potentially used for sorting
            item.Type = select(6, GetItemInfo(item.ID))
            self:Debug(i, item.ID, item.Count, item.Type)
            tinsert(guildspace.Bank, item)
        end

        guildspace.BankTimestamp = timestamp

        self:UpdateGuildBankGridView(1)
    else
        self:Print("error deserializing:", tbl)
        self:Print("Make sure both yours and the Guild Bank's", GHC.addon_name, "is up to date!")
    end
end
function GHC.Functions.ParseGuildBankData(...) return addon:ParseGuildBankData(...) end


function addon:UpdateGuildBankGridView(value)
    local guildspace = self:GetGuildspace()
    if not guildspace then return end

    if guildspace.Money then
        GHC.UI.GuildBankParentFrame.MoneyText:SetText(GetCoinTextureString(guildspace.Money))
    end

    if guildspace.Bank then
        -- table.sort(guildspace.Bank, function(a, b) return a.Type < b.Type end)

        GHC.UI.GuildBankScrollBar:SetMinMaxValues(1, math.max(math.ceil(#guildspace.Bank / 88), 1))

        -- Reset the entire guild bank grid view
        -- TODO Verify 96 is the right number...  might be 110 actually...?
        for i = 1, 96 do
            local gridViewItem = GHC.UI.GuildBankGridViewItems[i]
            gridViewItem.Texture:SetTexture(nil)
            gridViewItem.Text:SetText(nil)
            gridViewItem.Item = nil
            gridViewItem.Link = nil
        end

        -- Display the right portion of the bank based on scroll position
        -- TODO Verify 88 is the right number...  why is it 88?
        local v = math.ceil(value)
        local lBound = 88 * (v - 1) + 1
        local uBound = math.min(88 * v, #guildspace.Bank)
        for i = lBound, uBound do
            local item = guildspace.Bank[i]
            local gridViewItem = GHC.UI.GuildBankGridViewItems[i]
            gridViewItem.Texture:SetTexture(GetItemIcon(item.ID))
            gridViewItem.Text:SetText(item.Count)
            gridViewItem.Item = GetItemInfo(item.ID)
            gridViewItem.Link = select(2, GetItemInfo(item.ID))
        end
    end
end
function GHC.Functions.UpdateGuildBankGridView(...) return addon:UpdateGuildBankGridView(...) end


function addon:ShowGuildBankTooltip(frame, text)
    GameTooltip:SetOwner(frame) --, "ANCHOR_CURSOR")
    GameTooltip:SetPoint("TOPLEFT")
    if frame.Link then
        GameTooltip:SetHyperlink(frame.Link)
        GameTooltip:Show()
    end
end
function GHC.Functions.ShowGuildBankTooltip(...) return addon:ShowGuildBankTooltip(...) end


function GHC_SummaryFrameOnShow()

end


function GHC_BankFrameOnShow(self)
    addon:UpdateGuildBankGridView(1)
end


function GHC_RosterFrameOnShow()
    GHC.Var.ShouldUpdateDisplay = true
    addon:ScanGuildRoster()
end
function GHC_RosterFrameOnHide()
    GHC.Var.ShouldUpdateDisplay = false
end
-- addon:Print('loaded scripts')


-------------------------------------------------------------------------------
-- CORE
-------------------------------------------------------------------------------
function addon:PlayerDataCommHandler(prefix, message, distribution, sender)
    self:Debug("PlayerDataCommHandler", prefix, strlen(message), message, distribution, sender)
    self:ParsePlayerData(message, sender)
end


function addon:GuildBankMoneyCommHandler(prefix, message, distribution, sender)
    self:Debug("GuildBankMoneyCommHandler", prefix, strlen(message), message, distribution, sender)
    GHC.UI.GuildBankParentFrame.MoneyText:SetText(GetCoinTextureString(tostring(message)))
    -- TODO Rework this set of 3 functions (Start, Parse, Finish) to be more aptly named since
    -- AceComm handles packetizing large messages automatically
    self:GuildBankStart(message)
end


function addon:GuildBankDataCommHandler(prefix, message, distribution, sender)
    self:Debug("GuildBankDataCommHandler", prefix, strlen(message), message, distribution, sender)
    -- TODO Rework this set of 3 functions (Start, Parse, Finish) to be more aptly named since
    -- AceComm handles packetizing large messages automatically
    self:ParseGuildBankData(message, sender)
end


function addon:GuildBankFinishCommHandler(prefix, message, distribution, sender)
    self:Debug("GuildBankFinishCommHandler", prefix, strlen(message), message, distribution, sender)
    -- TODO Rework this set of 3 functions (Start, Parse, Finish) to be more aptly named
    self:GuildBankFinish()
end


function addon:UpdateMeCommHandler(prefix, message, distribution, sender)
    self:Debug("UpdateMeCommHandler", prefix, strlen(message), message, distribution, sender)
    if GetAddOnMetadata("GuildHelperClassic", "Version") ~= message then
        self:Print("New version available!")
        self.db.global.NewVersion = message
    end
end


--- Default Comm handler
function addon:OnCommReceived(prefix, message, distribution, sender)
    self:Debug("OnCommReceived", prefix, strlen(message), message, distribution, sender)
end


-- GetGuildInfo("player") should be available now
function addon:GUILD_ROSTER_UPDATE(eventName)
    self:Debug(eventName)

    -- Only need to register the Guildspace once

    if not IsGuildspaceRegistered then
        self:Load()
        IsGuildspaceRegistered = self:RegisterGuildspace(self:GetPlayerGuildInfo())
    end

    if self:GetGuildspace() then
        self:ScanGuildRoster()
    end
end


-- TODO Use the Ace Bucket module
function addon:PlayerDataChangedHandler(eventName, text)
    self:Debug(eventName, text)
    if self.db.global.AutoSendCharacterData then
        C_Timer.After(2, function() self:SendMyData() end)
    end
end

function addon:SKILL_LINES_CHANGED(eventName)
    self:Debug(eventName)
end


--- TODO Reset Guildspace and/or register new Guildspace
function addon:PLAYER_GUILD_UPDATE(eventName)
    local guildName = self:GetPlayerGuildInfo()
    if guildName then
    end
end


-- GetGuildInfo("player") definitely available now
-- TODO Is GetNumGuildMembers() available here??? If so this should be the event to use
-- for Registering Guildspace, etc.
function addon:GUILD_RANKS_UPDATE(eventName)
    self:Debug(eventName, UnitGUID("player"), GetGuildInfo("player"))
end


--- RegisterGuildspace can return true before GetNumGuildMembers() returns non-zero
-- RegisterGuildspace is lower level than GetGuildspace()
function addon:RegisterGuildspace(guildName)
    self:Debug("RegisterGuildspace", UnitGUID("player"), guildName, IsGuildspaceRegistered)
    if not guildName then return false end

    -- If the Guildspace db exists, then success, otherwise need to create new default table
    if self.db.factionrealm[guildName] then
        return true
    else
        local success = self:ResetGuildspace(guildName)
        if not success then return false end

        local newGuildspace = self.db.factionrealm[guildName]
        if not newGuildspace then return false end

        migrateOldGuildspace(self:GetPlayerGuildInfo())

        return true
    end
end


function addon:ResetGuildspace(guildName)
    self:Debug("ResetGuildspace", UnitGUID("player"), guildName)
    if not guildName then return false end

    self.db.factionrealm[guildName] = {
        Roster = {},
        RosterKeysSorted = {},
        RosterBackup = {},
        BankTimestamp = nil,
        Bank = {},
        Money = 0,
        ClassCount = {
            -- "DEATHKNIGHT" = { Count = 0 },
            DRUID = { Count = 0 },
            HUNTER = { Count = 0 },
            MAGE = { Count = 0 },
            PALADIN = { Count = 0 },
            PRIEST = { Count = 0 },
            ROGUE = { Count = 0 },
            SHAMAN = { Count = 0 },
            WARLOCK = { Count = 0 },
            WARRIOR = { Count = 0 },
        },
        ClassCountKeysSorted = {},
    }
    return true
end


function addon:GetGuildspace(guildName)
    if not IsGuildspaceRegistered then return nil end
    if GetNumGuildMembers() <= 0 then return nil end

    guildName = guildName or self:GetPlayerGuildInfo()
    self:Debug("GetGuildspace", UnitGUID("player"), guildName, IsGuildspaceRegistered)
    if not guildName then return nil end

    return self.db.factionrealm[guildName]
end


--- TODO Consider caching the value and only change if the player leaves or joins a guild (register new event)
function addon:GetPlayerGuildInfo()
    self:Debug("GetPlayerGuildInfo", IsInGuild("player"), GetGuildInfo("player"))
    return IsInGuild("player") and GetGuildInfo("player") or nil
end


function addon:ChatCommand(msg)
    if msg == nil or msg == "" then
        msg = "log"
    end
    local commands = {
        -- ["help"] = function() self:Print("Available commands are: open, scangb, fix, help") end,
        ["open"] = function() GHC.UI.Toggle() end,
        ["fix"] = function() StaticPopup_Show("GHC_FixMe") end,
        ["scangb"] = function() addon:ScanGuildBank() end,
        ["debug"] = function() DEBUG = not DEBUG; self:Print("DEBUG =", DEBUG) end,
    }
    local prefix, nextposition = self:GetArgs(msg, 1)
    local cmd = commands[prefix]
    if cmd ~= nil then
        cmd()
    else
        local message = self:GetArgs(msg, 1, nextposition)
        if DEBUG and prefix ~= nil and message ~= nil then
            self:Print("GHC_Slash: cmd:", prefix, message)
            self:SendCommMessage(prefix, message, "GUILD")
        else
            self:Print("Available commands are: open, scangb, fix, help")
        end
    end
end


function addon:Debug(...)
    if DEBUG then
        self:Print(...)
    end
end


--- Called 1 sec after addon loaded event allowing for GUIDs
function addon:Load()
    -- Create player settings db
    -- if not self.db.char then
    --     assign(
    --         self.db.char,
    --         createMemberTable(),
    --         {
    --             Name = UnitName("player"),
    --             Role = 0,
    --         }
    --     )
    -- end
    self.db.global.AutoSendCharacterData = self.db.global.AutoSendCharacterData or true
    self.db.global.CurrentVersion = self.db.global.CurrentVersion or GetAddOnMetadata("GuildHelperClassic", "Version")
    self.db.global.NewVersion = self.db.global.NewVersion or GHC.Var.NewVersion

    if self.db.global.AutoSendCharacterData then
        C_Timer.After(5, function() self:SendMyData() end)
    end

    self:AddonUpdater()

    -- local guildName, _, guildRankIndex, _ = GetGuildInfo("player")
    -- if guildRankIndex == 0 or guildRankIndex == 1 then
    --     GHC.UI.PlayerOptionsFrame.ScanGuildBankButton:Show()  -- does this work being located in the addon when you need to have bank open?
    -- end
end


-- TODO CurrentVersion is never updated
function addon:AddonUpdater()
    if self.db.global.NewVersion ~= self.db.global.CurrentVersion then
        StaticPopup_Show("GHC_UpdatePopup")
    end
end



--- This function is only called during the ADDON_LOADED event.
function addon:OnInitialize()
    self:RegisterChatCommand("ghc", "ChatCommand")

    self.db = LibStub("AceDB-3.0"):New("GuildHelperClassicDB", {
        global = {
            minimap = {
                hide = false,
            },
            AutoSendCharacterData = true,
        },
    })

    GHC.UI.MinimapIcon:Register(GHC.addon_name, GHC.UI.MinimapButton, self.db.global.minimap)
end


-- The OnEnable() and OnDisable() methods of your addon object are called by AceAddon when
-- your addon is enabled/disabled by the user.
-- Unlike OnInitialize(), this may occur multiple times without the entire UI being reloaded.

-- This function is only called during the PLAYER_LOGIN event, or during ADDON_LOADED,
-- if IsLoggedIn() already returns true at that point, e.g. for Load-on-Demand Addons.
--      IsLoggedIn() returns nil before the PLAYER_LOGIN event has fired, 1 afterwards.
function addon:OnEnable()
    -- restore saved settings
    -- register events
    -- register library callbacks
    -- check Saved Variables for frames' OnShow
    -- or other "on" setup
    GHC.UI.DrawGuildRosterListview()
    GHC.UI.DrawGuildClassSummaryBars()
    GHC.UI.DrawGuildBankGridView()
    -- GHC.UI.PlayerOptionsFrame.PlayerRoleDropdown_Init()
    GHC.UI.RosterScrollFrame.FilterProfDropDown_Init()
    GHC.UI.RosterScrollFrame.FilterRoleDropDown_Init()

    GHC.Functions.MakeFrameMove(GHC.UI.ParentFrame)

    self:RegisterComm("ghc-update-me", "UpdateMeCommHandler")
    self:RegisterComm("ghc-playerdata", "PlayerDataCommHandler")
    self:RegisterComm("ghc-gbdata-s", "GuildBankMoneyCommHandler")
    self:RegisterComm("ghc-gbdata", "GuildBankDataCommHandler")
    self:RegisterComm("ghc-gbdata-f", "GuildBankFinishCommHandler")

    self:RegisterEvent("GUILD_ROSTER_UPDATE")  -- GetGuildInfo("player") available
    self:RegisterEvent("GUILD_RANKS_UPDATE")  -- GetGuildInfo("player") definitely available
    -- self:RegisterEvent("PLAYER_GUILD_UPDATE")  -- Reset Guildspace

    self:RegisterEvent("SKILL_LINES_CHANGED")
    self:RegisterEvent("CHAT_MSG_SKILL", "PlayerDataChangedHandler")
    self:RegisterEvent("PLAYER_LEVEL_UP", "PlayerDataChangedHandler")
end


function addon:OnDisable()
    -- AceEvent-3.0 automatically handles unregistering events
end
